#include <float.h>

#include "genproc.h"
#include "shader_setup.h"

float ROUND(float f)
{
	if(f-floor(f) < 0.5)
		return floor(f);
	else
		return ceil(f);
}

// Constructor
GenProc::GenProc()
{
	texInternalFormat = GL_RGBA32F_ARB;
	npot_supported = false;
	distanceField = 0;
	attachmentpoints[0] = GL_COLOR_ATTACHMENT0_EXT;
	attachmentpoints[1] = GL_COLOR_ATTACHMENT1_EXT;
	writeTex = 0;
	readTex = 1;
	sh_prog_id_point = sh_prog_id_line = sh_prog_id_visualize = 0;
	texIDs[0] = texIDs[1] = 0;
	fbo = 0;
	fileLoaded = false;
}

// Destructor
GenProc::~GenProc()
{
	delete [] distanceField;
	cleanTextures();
	cleanShaderPrograms();
}

void GenProc::cleanTextures()
{
	glDeleteFramebuffersEXT( 1, &fbo );		// clean up framebuffer object
	glDeleteTextures( 2, texIDs );         // clean up textures
}

void GenProc::cleanShaderPrograms()
{
	glDeleteProgram(sh_prog_id_point);
	glDeleteProgram(sh_prog_id_line);
	glDeleteProgram(sh_prog_id_visualize);
}


bool GenProc::loadFile(const char *filename)
{
	points.clear();
	lines.clear();
	FILE *file = fopen(filename,"r");
	if(0 == file)
	{
		fprintf(stderr,"ERROR: Couldn't open file %s for reading\n",filename);
		return false;
	}

	fscanf(file,"distance-field\n");
	fscanf(file,"size %d %d\n",&fieldWidth,&fieldHeight);

	fscanf(file,"geometry\n");
	unsigned int numPoints;
	fscanf(file,"number of points %d\n",&numPoints);
	points.reserve(numPoints);
	for(unsigned int i = 0; i < numPoints; i++)
	{
		points.push_back(Point());
		fscanf(file,"%f %f %f %f %f\n",&(points[i].x),&(points[i].y),
			&(points[i].color[0]),&(points[i].color[1]),&(points[i].color[2]));
	}

	unsigned int numLines;
	fscanf(file,"number of lines %d\n",&numLines);
	lines.reserve(numLines);
	for(unsigned int i = 0; i < numLines; i++)
	{
		lines.push_back(Line());
		fscanf(file,"%f %f %f %f %f %f %f\n",&(lines[i].p[0]),&(lines[i].p[1]),&(lines[i].r[0]),&(lines[i].r[1]),
			&(lines[i].color[0]),&(lines[i].color[1]),&(lines[i].color[2]));
		// normalize direction vector, better now the later at each texel
		float norm = sqrt(lines[i].r[0]*lines[i].r[0] + lines[i].r[1]*lines[i].r[1]);
		lines[i].r[0] /= norm;
		lines[i].r[1] /= norm;
	}
	fileLoaded = true;
	reinit();
	return true;
}


void GenProc::loadShader()
{
	cleanShaderPrograms();
	initGLSL();	// init shader runtime
}

void GenProc::resetDistanceField()
{
	initDistanceField();
	transferToTexture( distanceField, texIDs[readTex] );
}

void GenProc::reinit()
{
	cleanTextures();

	const GLubyte *extstr = glGetString(GL_EXTENSIONS);
	npot_supported = (NULL != strstr(reinterpret_cast<const char *>(extstr),"GL_ARB_texture_non_power_of_two"));

	// check whether we can actually load textures of that size on the GPU
	glTexImage2D( GL_PROXY_TEXTURE_2D, 0, texInternalFormat,
		fieldWidth, fieldHeight, 0, GL_RGBA, GL_FLOAT, NULL );
	GLint realwidth;
	glGetTexLevelParameteriv( GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &realwidth );
	if ( realwidth == 0 )
	{
		fprintf( stderr, "Can't load textures of size %u x %u on the graphics card!\n",
			fieldWidth, fieldHeight );
		exit(1);
	}

	// check whether NPOT textures are supported (if any)
	if ( fieldWidth != ( 1 << static_cast<int> ( ROUND( log(static_cast<float>(fieldWidth))/log(2.0f) ) ) ) ||
		fieldHeight != ( 1 << static_cast<int> ( ROUND( log(static_cast<float>(fieldHeight))/log(2.0f) ) ) ) )
	{
		fprintf( stderr, "texture size is  %u x %u x RGBA\n", fieldWidth, fieldHeight );

		if ( npot_supported )
			fputs( "NPOT textures are OK.\n", stderr );
		else
		{
			fputs( "NPOT textures are NOT supported!\n", stderr );
			exit(1);
		}
	}

	initDistanceField();
	createTextures();					// create textures for vectors
	initFBO();							// init offscreen framebuffer
}

void GenProc::paintResult(const GLint windowWidth, const GLint windowHeight)
{
	glUseProgram(0); // switch back to fixed function pipeline
	// restore render destination to regular frame buffer
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );
	glViewport( 0, 0, windowWidth, windowHeight );
	glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	gluOrtho2D( 0.0, windowWidth, 0.0, windowHeight );

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClear(GL_COLOR_BUFFER_BIT);

	glActiveTexture( GL_TEXTURE0 );
	glEnable(GL_TEXTURE_2D);
	glBindTexture( GL_TEXTURE_2D, texIDs[readTex] );

	glPolygonMode( GL_FRONT, GL_FILL );
	glBegin(GL_QUADS);
	{
		glTexCoord2f(0.0, 0.0); glVertex2i(        0.0,          0.0);
		glTexCoord2f(1.0, 0.0); glVertex2i(windowWidth,          0.0);
		glTexCoord2f(1.0, 1.0); glVertex2i(windowWidth, windowHeight);
		glTexCoord2f(0.0, 1.0); glVertex2i(        0.0, windowHeight);
	}
	glEnd();
}


// initialize distance field with maximum available value
void GenProc::initDistanceField()
{
	delete [] distanceField;
	distanceField = new float[4*fieldWidth*fieldHeight];
	for(unsigned int i = 0; i < fieldWidth*fieldHeight; i++)
	{
		distanceField[4*i  ] = 0;
		distanceField[4*i+1] = 0;
		distanceField[4*i+2] = 0;
		distanceField[4*i+3] = FLT_MAX;
	}
}

/**
 * Create framebuffer object, bind it to reroute rendering operations 
 * from the traditional framebuffer to the off-screen buffer
 */
void GenProc::initFBO(void)
{
	// create FBO (off-screen framebuffer)
    glGenFramebuffersEXT( 1, &fbo ); 

    // bind offscreen framebuffer (that is, skip the window-specific render target)
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, fbo );

    // viewport for 1:1 pixel=texture mapping
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	gluOrtho2D( 0.0, fieldWidth, 0.0, fieldHeight );
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glViewport( 0, 0, fieldWidth, fieldHeight );

    // check if something went completely wrong
    checkGLErrors("initFBO()");
		// attach two textures to FBO
    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, attachmentpoints[writeTex],
							   GL_TEXTURE_2D, texIDs[writeTex], 0);
	checkGLErrors("performComputation() 0");
    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, attachmentpoints[readTex],
							   GL_TEXTURE_2D, texIDs[readTex], 0);
	checkGLErrors("initFBO() 2");

	// check if that worked
    if ( !checkFramebufferStatus() )
	{
		fputs( "glFramebufferTexture2DEXT() FAILED!\n", stderr );
		exit(1);
	}
}

void GenProc::createTextures(void)
{
    // create two textures, alternatingly read-only and write-only
    glGenTextures( 2, texIDs );

    // set up textures
    setupTexture( texIDs[readTex] );
	transferToTexture( distanceField, texIDs[readTex] );
    setupTexture( texIDs[writeTex] );

	// set texenv mode from modulate (the default) to replace)
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );

    // check if something went completely wrong
    checkGLErrors("createTextures()");
}

/**
 * Sets up a floating point texture with NEAREST filtering.
 * (mipmaps etc. are unsupported for floating point textures)
 */
void GenProc::setupTexture( const GLuint texID )
{
	// make active and bind
	glBindTexture( GL_TEXTURE_2D, texID );

	// turn off filtering and wrap modes!
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	// define texture with floating point format
	glTexImage2D( GL_TEXTURE_2D, 0, texInternalFormat,
				  fieldWidth, fieldHeight, 0, GL_RGBA, GL_FLOAT, NULL );

	// check if that worked
	checkGLErrors("setupTexture");
}


/**
 * Transfers data to texture. 
 * Check web page for detailed explanation on the difference between ATI and NVIDIA.
 */
void GenProc::transferToTexture( float* data, GLuint texID )
{
	// version (a): HW-accelerated on NVIDIA 
	glBindTexture( GL_TEXTURE_2D, texID);
	glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, fieldWidth, fieldHeight,
					 GL_RGBA, GL_FLOAT, data );

	// version (b): HW-accelerated on ATI 
//	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, texID, 0);
//	glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
//	glRasterPos2i(0,0);
//	glDrawPixels(texSize,texSize,GL_RGBA,GL_FLOAT,data);
//	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, 0, 0);
}


/**
 * Transfers data from currently texture, and stores it in given array.
 */
void GenProc::transferFromTexture(float* data)
{
	// version a: texture is attached
	// recommended on both NVIDIA and ATI
    glReadBuffer( attachmentpoints[readTex] );
	glReadPixels( 0, 0, fieldWidth, fieldHeight, GL_RGBA, GL_FLOAT, data );

	// version b: texture is not neccessarily attached
//	glBindTexture(GL_TEXTURE_2D,yTexID[readTex]);
//	glGetTexImage(GL_TEXTURE_2D,0,GL_RGBA,GL_FLOAT,data);
}


/**
 * Sets up the GLSL runtime and creates shader.
 */
void GenProc::initGLSL(void)
{
	// load and compile shader
	sh_prog_id_point = setShaders( "standard.vert", "distfield_point.frag" );
	// Get location of the texture samplers and point vector for future use
	texPointUni = glGetUniformLocation( sh_prog_id_point, "texture" );
	pointUni = glGetUniformLocation( sh_prog_id_point, "point" );
	pointColorUni = glGetUniformLocation( sh_prog_id_point, "pointColor" );

	sh_prog_id_line = setShaders( "standard.vert", "distfield_line.frag" );
	// Get location of the texture samplers for future use
	texLineUni = glGetUniformLocation( sh_prog_id_line, "texture" );
	lineUni = glGetUniformLocation( sh_prog_id_line, "linecoeff" );
	lineColorUni  = glGetUniformLocation( sh_prog_id_line, "lineColor" );

	sh_prog_id_visualize = setShaders( "standard.vert", "visualizeDF.frag" );
	texSizeRcpUni = glGetUniformLocation( sh_prog_id_visualize, "maxDistanceRCP" );
	texVisUni =  glGetUniformLocation( sh_prog_id_visualize, "texture" );
}

/**
 * Performs the actual calculation.
 */
void GenProc::performComputation(void)
{
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, fbo );
	glViewport( 0, 0, fieldWidth, fieldHeight );
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	gluOrtho2D( 0.0, fieldWidth, 0.0, fieldHeight );
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

	// process points
	glUseProgram( sh_prog_id_point );
	glUniform1i( texPointUni, 0 );
	for ( unsigned int i = 0; i < points.size(); i++ )
	{
		// set point coordinates for access in fragment program
		glUniform2f( pointUni, points[i].x, points[i].y );
		glUniform3fv( pointColorUni, 1, points[i].color );
		computationStep();
    }
	// process lines
	glUseProgram( sh_prog_id_line );
	glUniform1i( texLineUni, 0 );
	for ( unsigned int i = 0; i < lines.size(); i++ )
	{
		// set point coordinates for access in fragment program
		glUniform4f( lineUni, lines[i].p[0], lines[i].p[1], lines[i].r[0], lines[i].r[1] );
		glUniform3fv( lineColorUni, 1, lines[i].color );
		computationStep();
    }
	// done, just do some checks if everything went smoothly.
    checkFramebufferStatus();

	// ** DEBUGGING
	// print computed distance field to file
	//printResults("debug.txt");
	// ** DEBUGGING END

	// prepare texture for visualization
	glUseProgram(sh_prog_id_visualize);
	glUniform1i( texVisUni, 0 );
	float normalizef = 1.0f/sqrt(static_cast<float>(fieldWidth*fieldWidth + fieldHeight*fieldHeight));
	glUniform1f( texSizeRcpUni, normalizef );
	computationStep();
    checkGLErrors("performComputation()");

	//printResults("debug.txt");
}

void GenProc::computationStep()
{
	// set render destination
	glDrawBuffer( attachmentpoints[writeTex] );
	// set read texture
	glBindTexture( GL_TEXTURE_2D, texIDs[readTex] );
	// now render multitextured viewport-sized quad

	// make quad filled to hit every pixel/texel 
	// (should be default but we never know)
	glPolygonMode( GL_FRONT, GL_FILL );

	// and render the quad with normalized texcoords
	glBegin(GL_QUADS);
	{
		glTexCoord2f(0.0, 0.0); glVertex2f(       0.0,         0.0);
		glTexCoord2f(1.0, 0.0); glVertex2f(fieldWidth,         0.0);
		glTexCoord2f(1.0, 1.0); glVertex2f(fieldWidth, fieldHeight);
		glTexCoord2f(0.0, 1.0); glVertex2f(       0.0, fieldHeight);
	}
	glEnd();

	// swap role of the two textures (read-only source becomes 
	// write-only target and the other way round):
	swapTextures();
}

/**
 * swaps the role of the two y-textures (read-only and write-only)
 * Think of "pointer swapping".
 */
void GenProc::swapTextures(void)
{
    if ( writeTex == 0 )
	{
		writeTex = 1;
		readTex = 0;
    }
	else
	{
		writeTex = 0;
		readTex = 1;
    }
}

void GenProc::printResults(char *filename)
{
	FILE *file = fopen(filename,"w");
	if ( 0 == file) 
	{
		fprintf(stderr,"WARNING: Couldn't create file %s for writing computation results\n",filename);
		return;
	}
	transferFromTexture(distanceField);
	fprintf(file,"Distance Field\n");
	for(unsigned int i = 0; i < fieldHeight; i++)
	{
		for(unsigned int j = 0; j < fieldWidth; j++)
		{
			fprintf(file,"%.2f ",distanceField[4*(fieldWidth*i+j)+3]);
		}
		fprintf(file,"\n");
	}
	fprintf(file,"\nColors of Distance Field\n");
	for(unsigned int i = 0; i < fieldHeight; i++)
	{
		for(unsigned int j = 0; j < fieldWidth; j++)
		{
			fprintf(file,"(%.2f %.2f %.2f) ",distanceField[4*(fieldWidth*i+j)],distanceField[4*(fieldWidth*i+j)+1],distanceField[4*(fieldWidth*i+j)+2]);
		}
		fprintf(file,"\n");
	}

	fclose(file);
}


/**
 * Checks for OpenGL errors.
 * Extremely useful debugging function: When developing, 
 * make sure to call this after almost every GL call.
 */
void GenProc::checkGLErrors( const char *label )
{
    GLenum errCode = glGetError();
    if ( errCode == GL_NO_ERROR )
		return;

	fputs( "OpenGL ERROR: ", stderr);
	fputs( (char*)gluErrorString(errCode), stderr);
	fputs( " (label: ", stderr);
	fputs( label, stderr);
	fputs( ")\n", stderr);
}

/**
 * Checks framebuffer status.
 * Copied directly out of the spec, modified to deliver a return value.
 */
bool GenProc::checkFramebufferStatus(void)
{
    GLenum status;
    status = (GLenum) glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
    switch(status) {
        case GL_FRAMEBUFFER_COMPLETE_EXT:
            return true;
        case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
			printf("Framebuffer incomplete, incomplete attachment\n");
            return false;
        case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
			printf("Unsupported framebuffer format\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
			printf("Framebuffer incomplete, missing attachment\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
			printf("Framebuffer incomplete, attached images must have same dimensions\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
			printf("Framebuffer incomplete, attached images must have same format\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
			printf("Framebuffer incomplete, missing draw buffer\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
			printf("Framebuffer incomplete, missing read buffer\n");
            return false;
    }
	return false;
}
